#include <ncbi.h>
#include <gishlib.h>
#include "blastapp.h"

int LIBCALL
print_n(biop, hlp)
	BlastIoPtr	biop;
	BLAST_HitListPtr	hlp;
{
	FILE	*fp;
	unsigned long	id;
	BLAST_HSPPtr	hp;
	BLAST_StrPtr	dbsp;
	int		i, frame, prevframe = 0;
	BLAST_StrPtr	tstr;
	register char	*cp, *cpmax;

	fp = biop->fp;

	dbsp = &hlp->str2;
	id = dbsp->id.data.intvalue;

	db_seek(dbfp, id);
	db_get_seq(dbfp, dbsp);

	tstr = BlastStrNew(dbsp->len, nt4ap, 1);
	if (tstr == NULL)
		bfatal(ERR_UNDEF, "BlastStrNew failed:  %s", BlastErrStr(blast_errno));

	if (fp != NULL) {
		cp = dbsp->name;
		cpmax = cp + dbsp->namelen;
		while (cp < cpmax && !isspace(*cp))
			++cp;
		while (cp < cpmax && isspace(*cp))
			++cp;
		i = cp - dbsp->name + 1;
		i = MIN(i, 12);
		putc_unlocked('\n', fp);
		putc_unlocked('\n', fp);
		wrap(fp, ">", dbsp->name, dbsp->namelen, 79, i);
		fprintf(fp, "%*sLength = %s\n", i, "", Ltostr((long)dbsp->len,1));
	}

	cvt_str(dbsp, tstr);
	dbsp->str = tstr->str;
	dbsp->ap = tstr->ap;
	dbsp->len = dbsp->enclen = tstr->len;
	dbsp->lpb = 1;
	for (hp = *hlp->hpp; hp != NULL; hp = hp->next) {
		frame = hp->q_seg.frame;
		if (fp != NULL && SIGN(frame) != SIGN(prevframe)) {
			if (frame > 0)
				fprintf(fp, "\n  Plus");
			else
				fprintf(fp, "\n  Minus");
			fprintf(fp, " Strand HSPs:\n");
		}
		prevframe = frame;
		db_get_specific(dbfp, tstr, hp->s_seg.offset, hp->len);
		if (hsp_print(fp, hp, TRUE) != 0)
			return 1;
	}

	return 0;
}

int
cvt_str(src, dst)
	BLAST_StrPtr	src, dst;
{
	register BLAST_LetterPtr	sp, dp, dpmax, map;
	register BLAST_Letter	ch;
	static BLAST_AlphaMapPtr	amp;

	if (src->lpb != 4 || dst->lpb != 1)
		return 0;

	if (amp == NULL)
		amp = BlastAlphaMapFind(nt2ap, nt4ap);
	map = amp->map;

	dst->id = src->id;
	sp = src->str;  dp = dst->str;  dpmax = dp + src->len;
	*dpmax = nt4ap->sentinel;
	dst->frame = src->frame;
	dst->len = dst->enclen = src->len;
	dst->efflen = src->efflen;

	dpmax -= 4;
	while (dp < dpmax) {
		*dp++ = map[(ch = *sp++)>>6];
		*dp++ = map[(ch>>4)&3];
		*dp++ = map[(ch>>2)&3];
		*dp++ = map[ch&3];
	}
	dpmax += 4;
	while (dp < dpmax) {
		*dp++ = map[(ch = *sp++)>>6];
		if (dp >= dpmax)
			break;
		*dp++ = map[(ch>>4)&3];
		if (dp >= dpmax)
			break;
		*dp++ = map[(ch>>2)&3];
		if (dp >= dpmax)
			break;
		*dp++ = map[ch&3];
	}

	return 0;
}
